﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing;

namespace StateSpaceSearchCore
{
    class StateCostPair
    {
        public State State;
        public int Cost;

        public override int GetHashCode()
        {
            return State.GetHashCode() + Cost;
        }

        public StateCostPair(State s, int cost)
        {
            this.State = s;
            this.Cost = cost;
        }
    }

    abstract class State
    {
        public Problem p;
        public virtual bool IsFinal()
        {
            return p.IsFinal(this);
        }

        public abstract void GetSuccessors(List<StateCostPair> result);

        public abstract State Clone();
    }

    abstract class Problem
    {
        public abstract bool IsFinal(State s);

        public abstract void ReadFromFile(string filePath);

        public State InitialState;    
    }

    abstract class Visualizer
    {
        protected PictureBox screen;
        protected Graphics g;

        public virtual void Init(PictureBox p)
        {
            this.screen = p;
            screen.Image = new Bitmap(p.Width, p.Height);
            g = Graphics.FromImage(screen.Image);
        }

        protected void refresh()
        {
            screen.Refresh();
        }

        public abstract void Visualize(State s);
    }

}
